// graphics_xlib.cpp
// Revision 8-may-2006


#include "graphics_impl.h"

#include "config_threads.h"


#ifdef BLASSIC_USE_X


#include "util.h"
#include "key.h"
#include "thread.h"
#include "error.h"
#include "showerror.h"
#include "trace.h"


#include <string>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <vector>
#include <map>

#if defined __unix__ || defined __linux__
#include <unistd.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <sys/poll.h>


namespace blassic {

namespace graphics_impl {


using util::touch;
using util::to_string;

using std::string;
using std::cerr;
using std::endl;
using std::ostringstream;
using std::vector;
using std::map;
using std::make_pair;

using blassic::keynames::strCLICK;
using blassic::keynames::strSCLICK;
using blassic::keynames::strRELEASE;
using blassic::keynames::strSRELEASE;

static const long maskall= 0xFFFFFFFF;


typedef map <KeySym, string> mapkeyX_t;

mapkeyX_t mapkeyX;

inline void assignmapkeyX (KeySym code, const string & name)
{
	#ifndef NDEBUG
	mapkeyX_t::iterator it;
	if ( (it= mapkeyX.find (code) ) != mapkeyX.end () )
	{
		cerr << "Error: key " << name <<
			" has equal code than " << it->second <<
			endl;
		throw 1;
	}
	#endif
	mapkeyX [code]= name;
}

class InitMapkeyX {
	InitMapkeyX ();
	static InitMapkeyX instance;
};

InitMapkeyX InitMapkeyX::instance;

InitMapkeyX::InitMapkeyX ()
{
	TRF;

	using namespace blassic::keynames;

	assignmapkeyX (XK_Prior,     strPAGEUP);
	assignmapkeyX (XK_KP_Prior,  strPAGEUP);
	assignmapkeyX (XK_Next,      strPAGEDOWN);
	assignmapkeyX (XK_KP_Next,   strPAGEDOWN);
	assignmapkeyX (XK_End,       strEND);
	assignmapkeyX (XK_KP_End,    strEND);
	assignmapkeyX (XK_Home,      strHOME);
	assignmapkeyX (XK_KP_Home,   strHOME);
	assignmapkeyX (XK_Left,      strLEFT);
	assignmapkeyX (XK_KP_Left,   strLEFT);
	assignmapkeyX (XK_Up,        strUP);
	assignmapkeyX (XK_KP_Up,     strUP);
	assignmapkeyX (XK_Right,     strRIGHT);
	assignmapkeyX (XK_KP_Right,  strRIGHT);
	assignmapkeyX (XK_Down,      strDOWN);
	assignmapkeyX (XK_KP_Down,   strDOWN);
	assignmapkeyX (XK_Insert,    strINSERT);
	assignmapkeyX (XK_KP_Insert, strINSERT);
	assignmapkeyX (XK_Delete,    strDELETE);
	assignmapkeyX (XK_KP_Delete, strDELETE);
	assignmapkeyX (XK_F1,        strF1);
	assignmapkeyX (XK_F2,        strF2);
	assignmapkeyX (XK_F3,        strF3);
	assignmapkeyX (XK_F4,        strF4);
	assignmapkeyX (XK_F5,        strF5);
	assignmapkeyX (XK_F6,        strF6);
	assignmapkeyX (XK_F7,        strF7);
	assignmapkeyX (XK_F8,        strF8);
	assignmapkeyX (XK_F9,        strF9);
	assignmapkeyX (XK_F10,       strF10);
	assignmapkeyX (XK_F11,       strF11);
	assignmapkeyX (XK_F12,       strF12);
}

std::string string_from_key_X (KeySym key)
{
	mapkeyX_t::const_iterator it= mapkeyX.find (key);
	if (it != mapkeyX.end () )
		return it->second;
	return std::string ();
}


class ImageXWin : public Image {
public:
	ImageXWin (int x1, int y1, int x2, int y2,
		Display * display, Pixmap pixmap);
	~ImageXWin ();
	XImage * getImage ();
	int getwidth () const;
	int getheight () const;
private:
	XImage * img;
	int width;
	int height;
};


ImageXWin::ImageXWin (int x1, int y1, int x2, int y2,
		Display * display, Pixmap pixmap)
{
	width= x2 - x1 + 1;
	height= y2 - y1 + 1;
	img= XGetImage (display, pixmap, x1, y1, width, height,
		AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrFunctionCall;
}

ImageXWin::~ImageXWin ()
{
	XDestroyImage (img);
}

XImage * ImageXWin::getImage ()
{
	return img;
}

int ImageXWin::getwidth () const
{
	return width;
}

int ImageXWin::getheight () const
{
	return height;
}


class InitThreadsX {
	class DoInit {
		friend class InitThreadsX;
		DoInit ()
		{
			//XInitThreads ();
		}
	};
	static DoInit init;
};

InitThreadsX::DoInit InitThreadsX::init;


class GraphicsSystemXWin : public GraphicsSystem,
		public InitThreadsX,
		public blassic::Thread {
public:
	GraphicsSystemXWin (const std::string & sysname);
	Screen * newScreen ();
	//void init_pixmap ();

	Display * getDisplay () const;
	XIM getXIM () const;

	void process_event (XEvent & x_event);
	void check_events ();
	void wait_event ();
private:
	void do_wait_event ();
	Display * display;
	//int screen;
	XIM xim;

	void threadproc ();
};


class PainterXWin;


class ScreenXWin : public Screen {
public:
	ScreenXWin (GraphicsSystemXWin & system_n);
	void raw_create (int width, int height, int rwidth, int rheight,
		RotateMode rotmode, InkNum paperink);
private:
	void destroy ();
	void flush ();
	void wait_mapping ();
	void show ();
public:
	void set_title (const std::string & title);

	void set_synchro (bool synchro_n);
	bool get_synchro () const;

	void raw_init_colors ();

	void wait_event ();

	void set_border (InkNum inknum);
	void set_border_width (int newwidth);

	void setink (InkNum inknum, const ColorRGB & rgb);

	void init_pixmap ();

	void update_window (int x, int y, int width, int height);
	void update_window ();

	void setdrawmode (int drawmodenum);

	void raw_copy_rectangle (int x1from, int y1from,
		int x2from, int y2from, int xto, int yto);

	void raw_invert_rectangle (int x1, int y1, int x2, int y2);
	void raw_rectangle (int x1, int y1, int x2, int y2);

	void direct_fill_rectangle (int x1, int y1, int x2, int y2);

	void setactiveink (InkNum inknum);

	void raw_plot (int x, int y);
	int raw_test (int x, int y);

	bool load_char_image (int x, int y, int xzoom, int yzoom,
		unsigned char (& ch) [8] );

	Image * raw_getImage (int x1, int y1, int x2, int y2);
	void raw_putImage (Image & image, int x, int y, int mode);

	Painter * raw_getPainter (InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy);
	
	void real_plot_points (ShortPoint point [], int npoints);
	void real_plot_wide_points (ShortPoint point [], int npoints,
		int widex, int widey,
		int xmin, int ymin, int xmax, int ymax);

	void raw_draw_line (int prevx, int prevy, int x, int y);

private:
	void check_events ();
public:
	QueueKey & getQueueKey ();

	int raw_xmouse () const;
	int raw_ymouse () const;

	int keypressed (int keynum);
	void ring_my_bell ();
private:
	GraphicsSystemXWin & system;
	Display * display;
	int screen;

	int realwidth, realheight;
	int border_width;

	RotateMode rotatemode;
	Window borderwindow;
	Window mainwindow;
	bool border_mapped;
	bool main_mapped;
	bool main_viewable;
	Pixmap pixmap;
	bool pixmap_created;
	GC gc;
	GC gcp;
	GC gcwin_copy;
	GC gcpix_copy;
	int drawmode;
	XIC xic;
	Atom atomProtocols;
	Atom atomTakeFocus;

	bool synchro_mode;

	static const long eventusedmask=
		StructureNotifyMask | ExposureMask |
		FocusChangeMask |
		KeyPressMask | KeyReleaseMask |
		ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask | EnterWindowMask;
	static const long eventusedmaskborder=
		StructureNotifyMask | ExposureMask |
		FocusChangeMask;
	long eventusedmaskactual;

	struct UsedXColor {
		XColor color;
		ColorRGB rgb;
	};
	UsedXColor defaultinks [PREDEFINED_INKS];
	typedef map <InkNum, UsedXColor> usedinks_t;
	usedinks_t usedinks;

	UsedXColor & mapcolor (InkNum color);
	UsedXColor & mapnewcolor (InkNum color);

	static const InkNum inknone= static_cast <InkNum> (-1);
	bool border_set;
	InkNum borderink;
	InkNum activeink;
	InkNum windowactiveink;
	InkNum pixmapactiveink;

	static const int drawmode_copy= GXcopy;
	static const int drawmode_xor= GXxor;
	static const int drawmode_and= GXand;
	static const int drawmode_or= GXor;
	static const int drawmode_invert= GXinvert;

	// Keyboard

	QueueKey queuekey;

	void pushkey (const std::string & strkey);

	static const unsigned int MAXKEYCODE= 255;
	static const unsigned int NOUS= 0xFF;
	static const unsigned int inkeytocode [MAXINKEYCODE + 1];
	vector <bool> keycode_pressed;

	void keycode_press (unsigned int keycode);
	void keycode_release (unsigned int keycode);
	void keypress (XKeyPressedEvent & xk);
	void keyrelease (XKeyReleasedEvent & xk);

	// Mouse pointer

	int xmousepos, ymousepos;

	void checkink ();

	// X events

	void event_map (XMapEvent & evmap);
	void event_unmap (XUnmapEvent & evunmap);
	void event_configure (XConfigureEvent & evconf);
	void event_expose (XExposeEvent & evexpose);
	void event_focus_in (XFocusChangeEvent & evfocus);
	void event_client_message (XClientMessageEvent & climes);
public:
	void process_event (XEvent & x_event);
private:
	void wait_X_event ();

	static int drawmodebynumber (int drawmodenum);

	// Xlib calls wrapper

	void myXFlush ();
	void myXDrawPoint (int x, int y);
	void myXDrawPoints (XPoint * xpoint, int npoints);
	void myXFillRectangle (int x, int y,
		unsigned int width, unsigned int height);
	void myXDrawRectangle (int x, int y,
		unsigned int width, unsigned int height);
	void myXFillRectangles (XRectangle * rectangles, int nrectangles);
	void myXDrawSegments (XSegment * seg, int nsegments);

	friend class PainterXWin;
};


#define PAINT_MODE 3
#define PAINT_SYNC

class PainterXWin : public Painter {
public:
	PainterXWin (ScreenXWin & screen_n,
		InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy);
	~PainterXWin ();
private:
	ScreenXWin & screen;
	Locker locker;
	int min_x, min_y;
	int width, height;
	unsigned long colorpaint, colorborder;

	#if PAINT_MODE == 1

	XImage * image;

	#elif PAINT_MODE == 3

	static const size_t Bloque= 32;
	size_t xb, yb;

	//std::vector <std::vector <XImage *> > mimages;
	std::vector <XImage *> vimages;

	#ifndef NDEBUG
	size_t blocks_used;
	#endif

	#endif

	int olddrawmode;
	#ifdef PAINT_SYNC
	bool oldsync;
	#endif
	int x_min, x_max, y_min, y_max;
	void plot (Point p);
	bool do_check (Point p);
};

PainterXWin::PainterXWin (ScreenXWin & screen_n,
		InkNum inkpaint, InkNum inkborder,
		int minx, int miny, int maxx, int maxy) :
	Painter (minx, miny, maxx, maxy),
	screen (screen_n),
	locker (screen_n.system),
	min_x (minx), min_y (miny)
{
	TRF;

	colorpaint= screen.mapcolor (inkpaint).color.pixel;
	colorborder= screen.mapcolor (inkborder).color.pixel;
	width= maxx - minx + 1;
	height= maxy - miny + 1;

	#if PAINT_MODE == 1

	image= XGetImage (screen.display, screen.pixmap,
		minx, miny, width, height,
		AllPlanes, XYPixmap);
	if (image == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	#elif PAINT_MODE == 3

	xb= (static_cast <size_t> (width) + Bloque - 1) / Bloque;
	yb= (static_cast <size_t> (height) + Bloque - 1) / Bloque;
	TRDEBS ("Blocks: " << xb << ", " << yb);

	//mimages.resize (xb);
	//for (size_t i= 0; i < xb; ++i)
	//	mimages [i].resize (yb);
	vimages.resize (xb * yb);

	#ifndef NDEBUG
	blocks_used= 0;
	#endif

	#endif

	screen.setactiveink (inkpaint);
	screen.checkink ();

	#ifdef PAINT_SYNC
	oldsync= screen.synchro_mode;
	screen.synchro_mode= true;
	#endif
	x_min= y_min= 65000;
	x_max= y_max= 0;
}

PainterXWin::~PainterXWin ()
{
	TRF;
	
	#if PAINT_MODE == 1

	XDestroyImage (image);

	#elif PAINT_MODE == 3

	TRDEBS ("Destroying " << blocks_used << " blocks");

	#if 0
	for (size_t i= 0; i < xb; ++i)
	{
		for (size_t j= 0; j < yb; ++j)
		{
			if (mimages [i] [j] )
			{
				TRDEB ("Destroying block");
				XDestroyImage (mimages [i] [j] );
				#ifndef NDEBUG
				--blocks_used;
				#endif
			}
		}
	}
	#else
	for (size_t i= 0; i < xb * yb; ++i)
	{
		if (vimages [i] )
		{
			TRDEB ("Destroying block");
			XDestroyImage (vimages [i] );
			#ifndef NDEBUG
			--blocks_used;
			#endif
		}
	}
	#endif

	#ifndef NDEBUG
	if (blocks_used != 0)
	{
		TRDEB ("Error managing blocks");
	}
	#endif

	#endif

	#ifdef PAINT_SYNC
	screen.synchro_mode= oldsync;
	if (! oldsync)
	{
		//int x= min_x + x_min;
		//int y= min_y + y_min;
		int x= x_min;
		int y= y_min;
		int width= x_max - x_min + 1;
		int height= y_max - y_min + 1;
		TRDEBS ("Updating: " << x << ", " << y <<
			", " << width << ", " << height);
		screen.update_window (x, y, width, height);
	}
	#endif
}

//#if 0
void PainterXWin::plot (Point p)
{
	int x= p.x;
	int y= p.y;
	screen.raw_plot (p.x, p.y);
	if (x < x_min) x_min= x;
	if (x > x_max) x_max= x;
	if (y < y_min) y_min= y;
	if (y > y_max) y_max= y;
}
//#endif

bool PainterXWin::do_check (Point p)
{
	TRF;

	#if PAINT_MODE == 1

	unsigned long color= XGetPixel (image, p.x - min_x, p.y - min_y);

	#elif PAINT_MODE == 2

	unsigned long color= 0;
	XImage * image= XGetImage (screen.display, screen.pixmap,
		p.x - min_x, p.y - min_y, 1, 1,
		AllPlanes, XYPixmap);
	if (image != NULL)
	{
		color= XGetPixel (image, 0, 0);
		XDestroyImage (image);
	}

	#elif PAINT_MODE == 3

	int x= p.x - min_x;
	int y= p.y -  min_y;
	int xbcur= x / Bloque;
	int ybcur= y / Bloque;

	//XImage * & image= mimages [xbcur] [ybcur];
	XImage * & image= vimages [xbcur + ybcur * xb];

	if (image == 0)
	{
		TRDEB ("Creating block");
		int xg1= xbcur * Bloque;
		int yg1= ybcur * Bloque;
		int xg2= xg1 + Bloque;
		int yg2= yg1 + Bloque;
		if (xg2 > width)
			xg2= width;
		if (yg2 > height)
			yg2= height;
		ASSERT (xg2 > xg1);
		ASSERT (yg2 > yg1);
		image= XGetImage (screen.display, screen.pixmap,
			min_x + xg1, min_y + yg1, xg2 - xg1, yg2 - yg1,
			AllPlanes, XYPixmap);
		#ifndef NDEBUG
		++blocks_used;
		#endif
	}
	unsigned long color= XGetPixel (image, x % Bloque, y % Bloque);

	#endif

	bool r=  ! (color == colorpaint || color == colorborder);
	#if 0
	if (r)
	{
		int x= p.x;
		int y= p.y;
		screen.raw_plot (p.x, p.y);
		if (x < x_min) x_min= x;
		if (x > x_max) x_max= x;
		if (y < y_min) y_min= y;
		if (y > y_max) y_max= y;
	}
	#endif
	return r;
}

typedef map <Window, ScreenXWin *> winscreen_t;

static winscreen_t winscreen;


GraphicsSystemXWin::GraphicsSystemXWin (const std::string & sysname) :
	display (0),
	xim (0)
{
	TRF;

	const bool showfailinfo= showdebuginfo ();
	//const char * strDisplay;
	static const char WITHOUT_GRAPHICS []=
		", running without graphics support.";

	//if ( (strDisplay= getenv ("DISPLAY") ) != NULL &&
	//	strDisplay [0] != '\0')
	//{
		//TRMESSAGE (tr, string ("Opening ") + strDisplay);

		TRDEBS ("Opening " << sysname);

		char * displayname= const_cast <char *> (sysname.c_str () );
		display= XOpenDisplay (displayname);
		if (display)
		{
			//last_display= strDisplay;
			TRDEB ("Display opened");

			// Locale.

			#ifdef HAVE_LOCALE_H
			char * oldloc= setlocale (LC_ALL, NULL);
			setlocale (LC_ALL, "");
			#endif

			if (XSupportsLocale () )
			{
				TRDEB ("Setting locale modifiers");
				if (XSetLocaleModifiers ("") == NULL)
				{
					static const char FAILED []=
						"XSetLocaleModifiers failed";
					TRDEB (FAILED);
					if (showfailinfo)
						cerr << FAILED << endl;
				}
			}
			else
			{
				static const char UNSUPPORTED []=
					"Locale not supported by this display";
				TRDEB ( UNSUPPORTED);
				if (showfailinfo)
					cerr << UNSUPPORTED << endl;
			}

			TRDEB ("Opening XIM");
			xim= XOpenIM (display, NULL, NULL, NULL);
			if (xim == NULL)
			{
				static const char NOXIM []= "Can't open XIM";
				TRDEB (NOXIM);
				if (showfailinfo)
					cerr << NOXIM << endl;
			}
			else
			{
				TRDEB ("XIM opened");
			}

			#ifdef HAVE_LOCALE_H
			setlocale (LC_ALL, oldloc);
			#endif
		}
		else
		{
			static const char ERROR_OPEN []=
				"Error opening DISPLAY '";
			TRDEBS (ERROR_OPEN << sysname << '\'');
			if (showfailinfo)
				cerr << ERROR_OPEN <<
					sysname << '\'' <<
					WITHOUT_GRAPHICS <<
					endl;
			throw ErrFunctionCall;
		}
	#if 0
	}
	else
	{
		const char * const message= strDisplay ? "Empty" : "No";
		static const char DISPLAY []= " DISPLAY value";
		TRMESSAGE (tr, string (message) + DISPLAY);
		if (showfailinfo)
			cerr << message << DISPLAY << WITHOUT_GRAPHICS <<
				endl;
	}
	#endif

	TRDEB ("Starting");
	start ();
}

Display * GraphicsSystemXWin::getDisplay () const
{
	return display;
}

XIM GraphicsSystemXWin::getXIM () const
{
	return xim;
}

Screen * GraphicsSystemXWin::newScreen ()
{
	TRF;

	return new ScreenXWin (* this);
}


void GraphicsSystemXWin::process_event (XEvent & x_event)
{
	Window w= x_event.xany.window;
	winscreen_t::iterator it= winscreen.find (w);
	if (it != winscreen.end () )
		it->second->process_event (x_event);
}

void GraphicsSystemXWin::check_events ()
{
	TRF;

	XEvent x_event;
	while (XCheckMaskEvent (display, maskall, & x_event) == True)
	{
		process_event (x_event);
	}
}

void GraphicsSystemXWin::do_wait_event ()
{
	XEvent x_event;

	//XMaskEvent (display, maskall, & x_event);

	pollfd pfd;
	pfd.fd= ConnectionNumber (display);
	pfd.events= POLLIN | POLLPRI;
	int r= poll (& pfd, 1, -1);
	if (r < 0)
	{
		// Interruption, or unexpected error.
		return;
	}

	//XNextEvent (display, & x_event);
	//process_event (x_event);

	//XLockDisplay (display);
	lock ();
	while (XCheckMaskEvent (display, maskall, & x_event) == True)
	{
		process_event (x_event);
	}
	unlock ();
	//XUnlockDisplay (display);
}

void GraphicsSystemXWin::wait_event ()
{
	TRF;

	#ifdef BLASSIC_HAS_THREADS

	//XLockDisplay (display);
	XFlush (display);
	//XUnlockDisplay (display);

	TRDEB ("Locking");
	lock ();

	TRDEB ("Waiting");
	waitcond ();

	TRDEB ("Unlocking");
	unlock ();

	#else

	XFlush (display);
	do_wait_event ();

	#endif
}

void GraphicsSystemXWin::threadproc ()
{
	for (;;)
	{
		do_wait_event ();
	}
}


ScreenXWin::ScreenXWin (GraphicsSystemXWin & system_n) :
	Screen (system_n),
	system (system_n),
	display (0),
	screen (0),
	//border_width (15),
	border_width (0),
	borderwindow (0),
	mainwindow (0),
	border_mapped (false),
	main_mapped (false),
	main_viewable (false),
	pixmap (0),
	pixmap_created (false),
	drawmode (drawmode_copy),
	xic (0),
	atomProtocols (None),
	atomTakeFocus (None),
	synchro_mode (false),
	border_set (false),
	keycode_pressed (MAXKEYCODE + 1),
	xmousepos (0),
	ymousepos (0)
{
	TRF;
}

void ScreenXWin::raw_create (int width, int height, int rwidth, int rheight,
	RotateMode rotmode, InkNum paperink)
{
	TRF;

	Locker locker (system);

	const bool showfailinfo= showdebuginfo ();

	touch (width, height);
	realwidth= rwidth;
	realheight= rheight;
	rotatemode= rotmode;

	//border_width= 15;
	//borderink= paperink;

	activeink= paperink;
	windowactiveink= inknone;
	pixmapactiveink= inknone;

	display= system.getDisplay ();
	screen= DefaultScreen (display);
	Window rootwindow= RootWindow (display, screen);

	#if 0

	mainwindow= XCreateSimpleWindow (display,
		rootwindow,
		0, 0,
		realwidth, realheight,
		0,
		BlackPixel (display, screen),
		WhitePixel (display, screen) );

	unsigned int depth;
	{
		XWindowAttributes attr;
		XGetWindowAttributes (display, mainwindow, & attr);
		depth= attr.depth;
	}

	#else

	int depth= DefaultDepth (display, screen);

	#if 0
	mainwindow= XCreateWindow (display,
		rootwindow,
		0, 0,
		realwidth, realheight,
		0,
		depth,
		InputOutput,
		CopyFromParent,
		0,
		NULL);
	#endif

	int wbwidth= realwidth + border_width * 2;
	int wbheight= realheight + border_width * 2;

	borderwindow= XCreateWindow (display,
		rootwindow,
		0, 0,
		wbwidth, wbheight,
		0,
		depth,
		InputOutput,
		CopyFromParent,
		0,
		NULL);

	XSetWindowAttributes attr;
	attr.win_gravity= CenterGravity;
	mainwindow= XCreateWindow (display,
		borderwindow,
		border_width, border_width,
		realwidth, realheight,
		0,
		depth,
		InputOutput,
		CopyFromParent,
		CWWinGravity,
		& attr);

	#endif

	pixmap= XCreatePixmap (display, mainwindow,
		realwidth, realheight, depth);
	pixmap_created= true;

	XGCValues gcval;
	gcval.function= drawmode_copy;
	const unsigned long gcmask= GCFunction;

	gc= XCreateGC (display, mainwindow, gcmask, & gcval);
	gcp= XCreateGC (display, pixmap, gcmask, & gcval);

	gcwin_copy= XCreateGC (display, mainwindow, gcmask, & gcval);
	//XSetFunction (display, gc_copy, drawmode_copy);

	gcpix_copy= XCreateGC (display, pixmap, gcmask, & gcval);
	//XSetFunction (display, gcpix_copy, drawmode_copy);

	init_colors ();
	init_pixmap ();

	//XSetWindowBackground (display, borderwindow,
	//	mapcolor (0).color.pixel);
	if (border_set)
		set_border (borderink);
	else
		set_border (paperink);

	set_title (default_title);

	XSizeHints sh;
	sh.flags= PBaseSize | PMinSize | PMaxSize;
	sh.base_width= wbwidth;
	sh.min_width= wbwidth;
	sh.max_width= wbwidth;
	sh.base_height= wbheight;
	sh.min_height= wbheight;
	sh.max_height= wbheight;

	XSetWMProperties (display, borderwindow,
		NULL, NULL,
		NULL, 0,
		& sh, NULL, NULL);

	atomProtocols= XInternAtom (display, "WM_PROTOCOLS", True);
	if (atomProtocols != None)
	{
		atomTakeFocus= XInternAtom (display, "WM_TAKE_FOCUS", True);
		if (atomTakeFocus != None)
		{
			XSetWMProtocols (display, borderwindow,
				& atomTakeFocus, 1);
		}
		else
		{
			TRDEB ("WM_TAKE_FOCUS does not exist");
		}
	}
	else
	{
		TRDEB ("WM_PROTOCOLS does not exist");
	}

	eventusedmaskactual= eventusedmask;

	// Locale.

	#ifdef HAVE_LOCALE_H
	char * oldloc= setlocale (LC_ALL, NULL);
	setlocale (LC_ALL, "");
	#endif

	if (XSupportsLocale () )
	{
		TRDEB ("Setting locale modifiers");
		if (XSetLocaleModifiers ("") == NULL)
		{
			static const char FAILED []=
				"XSetLocaleModifiers failed";
			TRDEB (FAILED);
			if (showfailinfo)
				cerr << FAILED << endl;
		}
	}
	else
	{
		static const char UNSUPPORTED []=
			"Locale not supported by this display";
		TRDEB (UNSUPPORTED);
		if (showfailinfo)
			cerr << UNSUPPORTED << endl;
	}

	XIM xim= system.getXIM ();
	if (xim)
	{
		XIMStyle input_style= XIMPreeditNothing | XIMStatusNothing;
		xic= XCreateIC (xim,
			XNInputStyle, input_style,
			XNClientWindow, mainwindow,
			//XNFocusWindow, mainwindow,
			NULL);
	}
	if (xic != NULL)
	{
		TRDEB ("XIC created");

		long filterevents;
		if (XGetICValues (xic,
				XNFilterEvents, & filterevents,
				NULL)
			== NULL);
		{
			eventusedmaskactual |= filterevents;
		}
	}
	else
	{
		TRDEB ("XIC not created");
	}

	winscreen [borderwindow]= this;
	winscreen [mainwindow]= this;

	XSelectInput (display, borderwindow, eventusedmaskborder);

	XSelectInput (display, mainwindow, eventusedmaskactual);

	if (xic)
		XSetICFocus (xic);

	#if 0
	XMapWindow (display, mainwindow);
	XMapWindow (display, borderwindow);

	// Wait for window mapping.
	{
		XEvent e;
		do
		{
			XNextEvent (display, & e);
		} while (e.type != MapNotify);
	}
	#endif

	#ifdef HAVE_LOCALE_H
	setlocale (LC_ALL, oldloc);
	#endif

	//blassic::idle ();
}

void ScreenXWin::destroy ()
{
	TRF;

	if (! borderwindow)
		return;

	//XDestroyWindow (display, mainwindow);
	XDestroyWindow (display, borderwindow);

	XFreePixmap (display, pixmap);
	pixmap_created= false;

	myXFlush ();
	check_events ();

	winscreen.erase (mainwindow);
	winscreen.erase (borderwindow);
	
	borderwindow= 0;
}

void ScreenXWin::flush ()
{
	XFlush (display);
}

void ScreenXWin::wait_mapping ()
{
	TRF;

	#if 0
	XEvent e;
	bool waitmain= true, waitborder= true;
	do
	{
		XNextEvent (display, & e);
		if (e.type == MapNotify)
		{
			Window w= e.xany.window;
			if (w == borderwindow)
				waitborder= false;
			else if (w == mainwindow)
				waitmain= false;
		}
		process_event (e);
	} while (waitmain || waitborder);
	#endif

	while (! border_mapped || ! main_mapped)
		system.wait_event ();
}

void ScreenXWin::show ()
{
	TRF;

	XFlush (display);
	check_events ();

	XMapWindow (display, mainwindow);
	XMapWindow (display, borderwindow);
	XFlush (display);

	wait_mapping ();

	// Dispatch all pending events.
	#if 0
	XEvent e;
	while (XCheckMaskEvent (display, maskall, & e) == True)
		process_event (e);
	#endif
	check_events ();
}

void ScreenXWin::set_title (const std::string & title)
{
	char * aux= const_cast <char *> (title.c_str () );
	XTextProperty winname, iconname;
	XStringListToTextProperty (& aux, 1, & winname);
	XStringListToTextProperty (& aux, 1, & iconname);
	XSetWMProperties (display, borderwindow,
		& winname, & iconname,
		NULL, 0,
		NULL, NULL, NULL);
	XFree (winname.value);
	XFree (iconname.value);
}

void ScreenXWin::set_synchro (bool synchro_n)
{
	synchro_mode= synchro_n;
	if (synchro_mode)
		myXFlush ();
	else
	{
		update_window ();
		XSetFunction (display, gc, drawmode);
		XColor activecolor= mapcolor (activeink).color;
		XSetForeground (display, gc, activecolor.pixel);
	}
}

bool ScreenXWin::get_synchro () const
{
	return synchro_mode;
}

ScreenXWin::UsedXColor & ScreenXWin::mapcolor (InkNum color)
{
	if (color < PREDEFINED_INKS)
		return defaultinks [color];

	usedinks_t::iterator it= usedinks.find (color);
	if (it != usedinks.end () )
		return it->second;
	else
		return defaultinks [0];
}

ScreenXWin::UsedXColor & ScreenXWin::mapnewcolor (InkNum color)
{
	if (color < PREDEFINED_INKS)
		return defaultinks [color];

	usedinks_t::iterator it= usedinks.find (color);
	if (it != usedinks.end () )
		return it->second;
	else
	{
		UsedXColor n;
		return usedinks.insert (make_pair (color, n) ).first->second;
	}
}

void ScreenXWin::raw_init_colors ()
{
	TRF;

	XColor activecolor= mapcolor (get_graphics_pen () ).color;
	XSetForeground (display, gcp, activecolor.pixel);
	XSetForeground (display, gc, activecolor.pixel);
}

void ScreenXWin::setink (InkNum inknum, const ColorRGB & rgb)
{
	TRF;

	UsedXColor & ciu= mapnewcolor (inknum);

	ciu.rgb= rgb;
	Colormap cm= DefaultColormap (display, screen);
	XColor xc;

	ostringstream namecolor;
	namecolor << "rgb:" << std::hex << std::setfill ('0') <<
		std::setw (2) << rgb.r << '/' <<
		std::setw (2) << rgb.g << '/' <<
		std::setw (2) << rgb.b;
	XColor newpen;
	XAllocNamedColor (display, cm,
		namecolor.str ().c_str (), & newpen, & xc);

	// Not sure if previous color needs to be freed and why.
	ciu.color= newpen;

	if (inknum == pixmapactiveink)
		pixmapactiveink= inknone;
	if (inknum == windowactiveink)
		windowactiveink= inknone;
}

void ScreenXWin::set_border (InkNum inknum)
{
	TRF;

	borderink= inknum;
	border_set= true;
	if (! borderwindow)
		return;

	unsigned long pixel= mapcolor (borderink).color.pixel;
	XSetWindowBorder (display, borderwindow, pixel);
	XSetWindowBackground (display, borderwindow, pixel);
	XClearWindow (display, borderwindow);
}

void ScreenXWin::set_border_width (int newwidth)
{
	TRF;

	if (border_width == newwidth)
		return;

	border_width= newwidth;

	if (! borderwindow)
		return;

	//main_viewable= false;
	//border_mapped= false;
	//main_mapped= false;

	//XUnmapWindow (display, borderwindow);
	//XUnmapWindow (display, mainwindow);

	#if 0

	XResizeWindow (display, borderwindow,
		realwidth + border_width * 2,
		realheight + border_width * 2);

	#else

	//unsigned int mask= CWWidth | CWHeight;
	//XWindowChanges values;
	//values.width= realwidth + border_width * 2;
	//values.height= realheight + border_width * 2;
	//XReconfigureWMWindow (display, borderwindow, screen,
	//	mask, & values);

	int wbwidth= realwidth + border_width * 2;
	int wbheight= realheight + border_width * 2;
	XSizeHints sh;
	sh.flags= PBaseSize | PMinSize | PMaxSize;
	sh.base_width= wbwidth;
	sh.min_width= wbwidth;
	sh.max_width= wbwidth;
	sh.base_height= wbheight;
	sh.min_height= wbheight;
	sh.max_height= wbheight;

	XWMHints * phints= XAllocWMHints ();
	if (phints)
	{
		phints->flags= InputHint | StateHint;
		phints->input= True;
		phints->initial_state= NormalState;
	}

	XSetWMProperties (display, borderwindow,
		NULL, NULL,
		NULL, 0,
		& sh, phints, NULL);

	if (phints)
		XFree (phints);

	unsigned int mask= CWWidth | CWHeight;
	XWindowChanges values;
	values.width= wbwidth;
	values.height= wbheight;
	XReconfigureWMWindow (display, borderwindow, screen,
		mask, & values);

	#endif

	//XMoveWindow (display, mainwindow, border_width, border_width);

	//XMapWindow (display, borderwindow);
	//myXFlush ();
	//show ();
}

int ScreenXWin::drawmodebynumber (int drawmodenum)
{
	static const int drawmodesbynumber []= {
		drawmode_copy,
		drawmode_xor,
		drawmode_and,
		drawmode_or,
		drawmode_invert
	};

	if (drawmodenum < 0 || drawmodenum > drawmodenum_max)
		throw ErrFunctionCall;
	return drawmodesbynumber [drawmodenum];
}

void ScreenXWin::setdrawmode (int drawmodenum)
{
	drawmode= drawmodebynumber (drawmodenum);
	XSetFunction (display, gc, drawmode);
	if (! synchro_mode)
		XSetFunction (display, gcp, drawmode);
}

void ScreenXWin::checkink ()
{
	if (pixmapactiveink != activeink)
	{
		XColor activecolor= mapcolor (activeink).color;
		XSetForeground (display, gcp, activecolor.pixel);
		pixmapactiveink= activeink;
	}
	if (! synchro_mode && windowactiveink != activeink)
	{
		XColor activecolor= mapcolor (activeink).color;
		XSetForeground (display, gc, activecolor.pixel);
		windowactiveink= activeink;
	}
}

void ScreenXWin::setactiveink (InkNum inknum)
{
	TRF;

	activeink= inknum;
	//XColor activecolor= mapcolor (inknum).color;
	//XSetForeground (display, gcp, activecolor.pixel);
	//if (! synchro_mode)
	//	XSetForeground (display, gc, activecolor.pixel);
}

void ScreenXWin::init_pixmap ()
{
	XSetFunction (display, gcp, drawmode_copy);
	checkink ();
	XColor paper= mapcolor (activeink).color;
	XSetForeground (display, gcp, paper.pixel);
	XFillRectangle (display, pixmap, gcp, 0, 0, realwidth, realheight);
}

void ScreenXWin::update_window (int x, int y, int width, int height)
{
	//XSetFunction (display, gc, drawmode_copy);
	//XCopyArea (display, pixmap, mainwindow, gc,
	//	x, y, width, height, x, y);

	XCopyArea (display, pixmap, mainwindow, gcwin_copy,
		x, y, width, height, x, y);

	myXFlush ();
}

void ScreenXWin::update_window ()
{
	update_window (0, 0, realwidth, realheight);
}

void ScreenXWin::raw_copy_rectangle (int x1from, int y1from,
	int x2from, int y2from, int xto, int yto)
{
	TRFDEBS ("x1from: " << x1from << " y1from: " << y1from <<
		" x2from: " << x2from << " y2from: " << y2from <<
		" xto: " << xto << " yto: " << yto);

	int w= x2from - x1from + 1;
	int h= y2from - y1from + 1;

	#if 0
	XSetFunction (display, gcp, drawmode_copy);
	XCopyArea (display, pixmap, pixmap, gcp,
		x1from, y1from, w, h, xto, yto);
	XSetFunction (display, gcp, drawmode);
	#endif

	XCopyArea (display, pixmap, pixmap, gcpix_copy,
		x1from, y1from, w, h, xto, yto);

	if (! synchro_mode)
	{
		// Update just the copied to area.
		#if 0
		XSetFunction (display, gc, drawmode_copy);
		XCopyArea (display, pixmap, mainwindow, gc,
			xto, yto, w, h, xto, yto);
		XSetFunction (display, gc, drawmode);
		#endif
		update_window (xto, yto, xto + w, yto + h);
	}
}

void ScreenXWin::raw_invert_rectangle (int x1, int y1, int x2, int y2)
{
	XSetFunction (display, gc, drawmode_invert);
	if (! synchro_mode)
		XSetFunction (display, gcp, drawmode_invert);

	checkink ();
	myXFillRectangle (x1, y1, x2 - x1, y2 - y1);

	XSetFunction (display, gc, drawmode);
	if (! synchro_mode)
	{
		XSetFunction (display, gcp, drawmode);
	}
}

void ScreenXWin::raw_rectangle (int x1, int y1, int x2, int y2)
{
	checkink ();
	myXDrawRectangle (x1, y1, x2 - x1, y2 - y1);	
}

void ScreenXWin::direct_fill_rectangle (int x1, int y1, int x2, int y2)
{
	TRF;

	int w= x2 - x1 + 1;
	int h= y2 - y1 + 1;

	TRDEBS ("x1: " << x1 << " x2: " << x2 << " w: " << w << " h: " << h);

	checkink ();
	myXFillRectangle (x1, y1, w, h);
}

void ScreenXWin::raw_plot (int x, int y)
{
	checkink ();
	myXDrawPoint (x, y);
}

int ScreenXWin::raw_test (int x, int y)
{
	XImage * img= XGetImage (display, pixmap, x, y, 1, 1,
		AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	unsigned long color= XGetPixel (img, 0, 0);
	XDestroyImage (img);
	for (int i= 0; i < 16; ++i)
		if (mapcolor (i).color.pixel == color)
			return i;
	for (usedinks_t::iterator it= usedinks.begin ();
		it !=usedinks.end (); ++it)
	{
		if (it->second.color.pixel == color)
			return it->first;
	}
	return -1;
}

bool ScreenXWin::load_char_image (int x, int y, int xzoom, int yzoom,
	unsigned char (& ch) [8] )
{
	TRFDEBS ("x: " << x << " y: " << y);

	class ImageGuard {
	public:
		ImageGuard (XImage * img) :
			img (img)
		{ }
		~ImageGuard ()
		{
			XDestroyImage (img);
		}
	private:
		XImage * img;
	};

	int width= 8 * xzoom;
	int height= 8 * yzoom;
	XImage * img= XGetImage (display, pixmap, x, y,
		width, height, AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	ImageGuard guard (img);
	unsigned long back= XGetPixel (img, 0, 0);
	bool fFore= false;
	unsigned long fore= 0;
	for (int i= 0, ipos= 0; i < 8; ++i, ipos+= yzoom)
		for (int j= 0, jpos= 0; j < 8; ++j, jpos+= xzoom)
		{
			unsigned long c= XGetPixel (img, jpos, ipos);
			bool bit;
			if (c == back)
			{
				bit= false;
			}
			else
			{
				if (! fFore)
				{
					fFore= true;
					fore= c;
				}
				if (c != fore)
					return false;
				bit= true;
			}
			if (xzoom > 1 || yzoom > 1)
			{
				const int iend= ipos + yzoom;
				const int jend= jpos + xzoom;
				for (int ii= ipos; ii < iend; ++ii)
					for (int jj= jpos; jj < jend; ++jj)
					{
						if (XGetPixel (img, jj, ii)
								!= c)
							return false;
					}
			}
			ch [i]<<= 1;
			ch [i]|= bit;
		}
	return true;
}

Image * ScreenXWin::raw_getImage (int x1, int y1, int x2, int y2)
{
	TRF;

	return new ImageXWin (x1, y1, x2, y2, display, pixmap);
}

void ScreenXWin::raw_putImage (Image & image, int x, int y, int mode)
{
	TRF;

	ImageXWin * pimageX= dynamic_cast <ImageXWin *> (& image);
	if (pimageX == NULL)
		throw ErrFunctionCall;

	int maxwidth= realwidth - x + 1;
	int maxheight= realheight - y + 1;
	int w= std::min (pimageX->getwidth (), maxwidth);
	int h= std::min (pimageX->getheight (), maxheight);

	int modeused= drawmodebynumber (mode);

	XImage * img= pimageX->getImage ();

	XSetFunction (display, gcp, modeused);
	XPutImage (display, pixmap, gcp, img,
		0, 0, x, y, w, h);
	XSetFunction (display, gcp, drawmode);

	if (! synchro_mode)
		update_window (x, y, x + w, y + h);
}

Painter * ScreenXWin::raw_getPainter (InkNum inkpaint, InkNum inkborder,
	int minx, int miny, int maxx, int maxy)
{
	return new PainterXWin (* this, inkpaint, inkborder, minx, miny, maxx, maxy);
}


void ScreenXWin::real_plot_points (ShortPoint point [], int npoints)
{
	TRF;

	XPoint * xpoint= reinterpret_cast <XPoint *> (point);

	checkink ();
	myXDrawPoints (xpoint, npoints);
}

void ScreenXWin::real_plot_wide_points (ShortPoint point [], int npoints,
	int widex, int widey,
	int xmin, int ymin, int xmax, int ymax)
{
	TRFDEBS ("widex: " << widex << " widey: " << widey);

	checkink ();
	if (widex == 1 || widey == 1)
	{
		if (widex == 1)
		{
			widex= 0;
			--widey;
		}
		else
		{
			widey= 0;
			--widex;
		}

		XPoint * xpoint= reinterpret_cast <XPoint *> (point);

		// 64 points each char
		XSegment seg [64];

		for (int i= 0; i < npoints; ++i)
		{
			int xpos= xpoint [i].x;
			int ypos= xpoint [i].y;
			seg [i].x1= std::max (xpos, xmin);
			seg [i].y1= std::max (ypos, ymin);
			xpos+= widex;
			ypos+= widey;
			seg [i].x2= std::min (xpos, xmax);
			seg [i].y2= std::min (ypos, ymax);
		}
		myXDrawSegments (seg, npoints);
	}
	else
	{
		XRectangle rect [64];
		for (int i= 0; i < npoints; ++i)
		{
			int x= point [i].x;
			int y= point [i].y;
			int x1= std::max (x, xmin);
			int y1= std::max (y, ymin);
			rect [i].x= x1;
			rect [i].y= y1;
			int x2= std::min (x + widex - 1, xmax);
			int y2= std::min (y + widey - 1, ymax);
			rect [i].width= x2 - x1 + 1;
			rect [i].height= y2 - y1 + 1;
		}
		myXFillRectangles (rect, npoints);
	}
}

void ScreenXWin::raw_draw_line (int prevx, int prevy, int x, int y)
{
	checkink ();
	XDrawLine (display, pixmap, gcp, prevx, prevy, x, y);
	if (! synchro_mode)
		XDrawLine (display, mainwindow, gc, prevx, prevy, x, y);
	myXFlush ();
}

QueueKey & ScreenXWin::getQueueKey ()
{
	return queuekey;
}

const unsigned int ScreenXWin::inkeytocode [MAXINKEYCODE + 1]= {
	// The key symbols indicated are the corresponding
	// to a spanish keyboard.
	0x62, //  0 Up
	0x66, //  1 Right
	0x68, //  2 Down
	0x51, //  3 Numeric 9
	0x55, //  4 Numeric 6
	0x59, //  5 Numeric 3
	0x6C, //  6 Intro
	0x5B, //  7 Numeric .
	0x64, //  8 Left
	0x00, //  9 Copy
	0x4F, // 10 Numeric 7
	0x50, // 11 Numeric 8
	0x54, // 12 Numeric 5
	0x57, // 13 Numeric 1
	0x58, // 14 Numeric 2
	0x5A, // 15 Numeric 0
	0x6B, // 16 Delete
	0x23, // 17 +
	0x24, // 18 Return
	0x33, // 19
	0x53, // 20 Numeric 4
	0x3E, // 21 Shift
	0x5E, // 22 \ in the CPC, <>
	0x6D, // 23 Control
	0x15, // 24
	0x14, // 25 '
	0x22, // 26 `[
	0x21, // 27 P
	0x30, // 28 '{
	0x2F, // 29
	0x3D, // 30 -
	0x3C, // 31 .
	0x13, // 32 0
	0x12, // 33 9
	0x20, // 34 O
	0x1F, // 35 I
	0x2E, // 36 L
	0x2D, // 37 K
	0x3A, // 38 M
	0x3B, // 39 ,
	0x11, // 40 8
	0x10, // 41 7
	0x1E, // 42 U
	0x1D, // 43 Y
	0x2B, // 44 H
	0x2C, // 45 J
	0x39, // 46 N
	0x41, // 47 space
	0x0F, // 48 6
	0x0E, // 49 5
	0x1B, // 50 R
	0x1C, // 51 T
	0x2A, // 52 G
	0x29, // 53 F
	0x38, // 54 B
	0x37, // 55 V
	0x0D, // 56 4
	0x0C, // 57 3
	0x1A, // 58 E
	0x19, // 59 W
	0x27, // 60 S
	0x28, // 61 D
	0x36, // 62 C
	0x35, // 63 X
	0x0A, // 64 1
	0x0B, // 65 2
	0x09, // 66 Escape
	0x18, // 67 Q
	0x17, // 68 Tab
	0x26, // 69 A
	0x42, // 70 Caps lock
	0x34, // 71 Z
	NOUS, // 72 joystick on cpc, unasigned
	NOUS, // 73 joystick on cpc, unasigned
	NOUS, // 74 joystick on cpc, unasigned
	NOUS, // 75 joystick on cpc, unasigned
	NOUS, // 76 joystick on cpc, unasigned
	NOUS, // 77 joystick on cpc, unasigned
	NOUS, // 78 Inexistent
	0x16, // 79 Backspace
};

void ScreenXWin::keycode_press (unsigned int keycode)
{
	//std::cerr << "Pressed " << std::hex << std::setfill ('0') <<
	//	std::setw (3) << keycode << std::dec << std::endl;
	if (keycode <= MAXKEYCODE)
		keycode_pressed [keycode]= true;
	else
	{
		if (showdebuginfo () )
			cerr << "X key code out of range: " << keycode << endl;
	}
}

void ScreenXWin::keycode_release (unsigned int keycode)
{
	//std::cerr << "Released " << std::hex << std::setfill ('0') <<
	//	std::setw (3) << keycode << std::dec << std::endl;
	if (keycode <= MAXKEYCODE)
		keycode_pressed [keycode]= false;
	else
	{
		if (showdebuginfo () )
			cerr << "X key code out of range: " << keycode << endl;
	}
}

int ScreenXWin::keypressed (int keynum)
{
	TRF;

	//cerr << "keypressed (" << keynum << ")" << endl;

	if (keynum < 0 || keynum > static_cast <int> (MAXINKEYCODE) )
	{
		if (showdebuginfo () )
			cerr << "INKEY code out of range: " << keynum << endl;
		return -1;
	}

	check_events ();

	const int shiftpressed= 32, ctrlpressed= 128;

	const unsigned int
		shift_left= 0x32, shift_right= 0x3E,
		control_left= 0x25, control_right= 0x6D;

	bool shifted= keycode_pressed [shift_left] ||
		keycode_pressed [shift_right];
	bool controled= keycode_pressed [control_left] ||
		keycode_pressed [control_right];

	bool pressed;
	switch (keynum)
	{
	case 21:
		pressed= shifted;
		break;
	case 23:
		pressed= controled;
		break;
	default:
		pressed= keycode_pressed [inkeytocode [keynum] ];
	}
	if (! pressed)
		return -1;
	int r= 0;
	if (shifted)
		r|= shiftpressed;
	if (controled)
		r|= ctrlpressed;

	return r;
}

void ScreenXWin::pushkey (const std::string & strkey)
{
	//system.lock ();
	queuekey.push (strkey);
	system.setcond ();
	//system.unlock ();
}

void ScreenXWin::keypress (XKeyPressedEvent & xk)
{
	//TRACEFUNC (tr, "ScreenXWin::keypress");

	KeySym ks= 0;
	const int STRBUFSIZE= 500; // Value used in xterm.
	char buffer [STRBUFSIZE];
	int lookupresult;
	Status lkstatus;
	if (xic != NULL)
	{
		lookupresult= XmbLookupString (xic, & xk, buffer, STRBUFSIZE,
			& ks, & lkstatus);
	}
	else
	{
		lookupresult= XLookupString (& xk, buffer, STRBUFSIZE - 1,
			& ks, NULL);
		lkstatus= lookupresult > 0 ? XLookupChars : XLookupKeySym;
	}

	// Change cursor keys if rotated.
	switch (rotatemode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		switch (ks)
		{
		case XK_Left:
			ks= XK_Up; break;
		case XK_Up:
			ks= XK_Right; break;
		case XK_Right:
			ks= XK_Down; break;
		case XK_Down:
			ks= XK_Left; break;

		case XK_KP_Left:
			ks= XK_KP_Up; break;
		case XK_KP_Up:
			ks= XK_KP_Right; break;
		case XK_KP_Right:
			ks= XK_KP_Down; break;
		case XK_KP_Down:
			ks= XK_KP_Left; break;
		}
		break;
	}

	switch (lkstatus)
	{
	case XBufferOverflow:
		// This is supposed to never happen.
		//TRMESSAGE (tr, "X keyboard overflow");
		break;
	case XLookupBoth:
	case XLookupKeySym:
		#ifdef XK_EuroSign
		// If the locale is not adequate the euro sign can
		// be not treated as character, this make it work.
		if (ks == XK_EuroSign)
		{
			//TRMESSAGE (tr, "Euro workaround used");
			const char eurochar (164);
			pushkey (string (1, eurochar) );
		}
		else
		#endif
		{
			string str= string_from_key_X (ks);
			if (! str.empty () )
				pushkey (str);
			else if (lkstatus == XLookupBoth)
				pushkey (string (buffer, lookupresult) );
		}
		break;
	case XLookupChars:
		pushkey (string (buffer, lookupresult) );
		break;
	default:
		break;
	}
}

void ScreenXWin::keyrelease (XKeyReleasedEvent & xk)
{
	KeySym ks= 0;
	const int STRBUFSIZE= 500; // Value used in xterm.
	char buffer [STRBUFSIZE];
	int r;
	if (xic == NULL)
	{
		r= XLookupString (& xk, buffer, STRBUFSIZE - 1,
			& ks, NULL);
	}

	// Change cursor keys if rotated.
	switch (rotatemode)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		switch (ks)
		{
		case XK_Left:
			ks= XK_Up; break;
		case XK_Up:
			ks= XK_Right; break;
		case XK_Right:
			ks= XK_Down; break;
		case XK_Down:
			ks= XK_Left; break;

		case XK_KP_Left:
			ks= XK_KP_Up; break;
		case XK_KP_Up:
			ks= XK_KP_Right; break;
		case XK_KP_Right:
			ks= XK_KP_Down; break;
		case XK_KP_Down:
			ks= XK_KP_Left; break;
		}
		break;
	}
	//reset_pressed (ks);
}

int ScreenXWin::raw_xmouse () const
{
	return xmousepos;
}

int ScreenXWin::raw_ymouse () const
{
	return ymousepos;
}

void ScreenXWin::ring_my_bell ()
{
	XBell (display, 100);
}

void ScreenXWin::event_map (XMapEvent & evmap)
{
	TRFDEBS ("Window: " <<
		( (evmap.window == borderwindow) ? "Border" : "Main") );

	if (evmap.window == borderwindow)
	{
		border_mapped= true;
		if (main_mapped)
			main_viewable= true;
	}
	else if (evmap.window == mainwindow)
	{
		main_mapped= true;
		if (border_mapped)
			main_viewable= true;
	}
	system.setcond ();
}

void ScreenXWin::event_unmap (XUnmapEvent & evunmap)
{
	TRFDEBS ("Window: " <<
		( (evunmap.window == borderwindow) ? "Border" : "Main") );

	if (evunmap.window == borderwindow)
	{
		border_mapped= false;
		main_viewable= false;
	}
	if (evunmap.window == mainwindow)
	{
		main_mapped= false;
		main_viewable= false;
	}
}

void ScreenXWin::event_configure (XConfigureEvent & evconf)
{
	TRFDEBS ("Window: " <<
		( (evconf.window == borderwindow) ? "Border" : "Main") <<
		" width: " << evconf.width << " height: " << evconf.height);

	touch (evconf);
}

void ScreenXWin::event_expose (XExposeEvent & evexpose)
{
	TRF;

	if (evexpose.window == mainwindow)
	{
		TRDEB ("Main window");

		int x1= evexpose.x;
		int x2= x1 + evexpose.width;
		int y1= evexpose.y;
		int y2= y1 + evexpose.height;
		update_window (x1, y1, x2, y2);
		if (! synchro_mode)
			XSetFunction (display, gc, drawmode);
	}
}

void ScreenXWin::event_focus_in (XFocusChangeEvent & evfocus)
{

	if (evfocus.window == borderwindow)
	{
		if (main_viewable)
		{
			XSetInputFocus (display, mainwindow,
				RevertToParent, CurrentTime);
		}
	}
}

void ScreenXWin::event_client_message (XClientMessageEvent & evclient)
{
	TRF;

	if (evclient.message_type == atomProtocols &&
		evclient.window == borderwindow &&
		static_cast <Atom> (evclient.data.l [0] ) == atomTakeFocus &&
		main_viewable)
	{
		TRDEB ("Setting focus");
		XSetInputFocus (display, mainwindow, RevertToParent,
			evclient.data.l [1] );
	}
}

void ScreenXWin::process_event (XEvent & x_event)
{
	// Set the state of pressed keys before checking XFilterEvent.
	switch (x_event.type)
	{
	case KeyPress:
		keycode_press (x_event.xkey.keycode);
		break;
	case KeyRelease:
		keycode_release (x_event.xkey.keycode);
		break;
	}

	if (XFilterEvent (& x_event, mainwindow) )
		return;

	switch (x_event.type)
	{
	case MapNotify:
		event_map (x_event.xmap);
		break;
	case UnmapNotify:
		event_unmap (x_event.xunmap);
		break;
	case ConfigureNotify:
		event_configure (x_event.xconfigure);
		break;
	case ClientMessage:
		event_client_message (x_event.xclient);
		break;
	case Expose:
		event_expose (x_event.xexpose);
		break;
	case FocusIn:
		event_focus_in (x_event.xfocus);
		break;
	case KeyPress:
		keypress (x_event.xkey);
		break;
	case KeyRelease:
		keyrelease (x_event.xkey);
		break;
	case ButtonPress:
		{
			XButtonEvent & xbpe= x_event.xbutton;
			//cerr << "ButtonPress event, button=" <<
			//	xbpe.button <<
			//	endl;
			switch (xbpe.button)
			{
			case 1:
				pushkey (strCLICK);
				break;
			case 3:
				pushkey (strSCLICK);
				break;
			default:
				;
			}
		}
		break;
	case ButtonRelease:
		{
			XButtonEvent & xbpe= x_event.xbutton;
			//cerr << "ButtonRelease event, button=" <<
			//	xbpe.button <<
			//	endl;
			switch (xbpe.button)
			{
			case 1:
				pushkey (strRELEASE);
				break;
			case 3:
				pushkey (strSRELEASE);
				break;
			default:
				;
			}
		}
		break;
	case MotionNotify:
		{
			XMotionEvent & xme= x_event.xmotion;
			//cerr << "MotionNotify event " <<
			//	xme.x << ", " << xme.y <<
			//	endl;
			xmousepos= xme.x;
			ymousepos= xme.y;
			//do_unrotate (xmousepos, ymousepos);
		}
		break;
	case EnterNotify:
		{
			XCrossingEvent  & xce= x_event.xcrossing;
			//cerr << "EnterNotify event" << endl;
			xmousepos= xce.x;
			ymousepos= xce.y;
			//do_unrotate (xmousepos, ymousepos);
		}
		break;
	default:
		//cerr << "Another event." << endl;
		;
	} // switch type of event
}

void ScreenXWin::wait_X_event ()
{
	XEvent x_event;

	//XWindowEvent (display, mainwindow, eventusedmaskactual, & x_event);
	XMaskEvent (display, maskall, & x_event);

	process_event (x_event);
}

void ScreenXWin::wait_event ()
{
	TRF;

	//wait_X_event ();
	system.wait_event ();
}

void ScreenXWin::check_events ()
{
	TRF;

	#if 0
	XEvent x_event;

	//while (XCheckWindowEvent (display, mainwindow,
	//	eventusedmaskactual, & x_event) )
	//	process_event (e);
	while (XCheckMaskEvent (display, maskall, & x_event) == True)
	{
		process_event (x_event);
	}
	#endif

	system.check_events ();
}

void ScreenXWin::myXFlush ()
{
	XFlush (display);
}

void ScreenXWin::myXDrawPoint (int x, int y)
{
	XDrawPoint (display, pixmap, gcp, x, y);
	if (! synchro_mode)
		XDrawPoint (display, mainwindow, gc, x, y);
}

void ScreenXWin::myXDrawPoints (XPoint * xpoint, int npoints)
{
	XDrawPoints (display, pixmap, gcp,
		xpoint, npoints, CoordModeOrigin);
	if (! synchro_mode)
		XDrawPoints (display, mainwindow, gc,
			xpoint, npoints, CoordModeOrigin);
}

void ScreenXWin::myXFillRectangle (int x, int y,
		unsigned int width, unsigned int height)
{
	XFillRectangle (display, pixmap, gcp, x, y, width, height);
	if (! synchro_mode)
		XFillRectangle (display, mainwindow, gc, x, y, width, height);
}

void ScreenXWin::myXDrawRectangle (int x, int y,
		unsigned int width, unsigned int height)
{
	XDrawRectangle (display, pixmap, gcp, x, y, width, height);
	if (! synchro_mode)
		XDrawRectangle (display, mainwindow, gc, x, y, width, height);
}

void ScreenXWin::myXFillRectangles (XRectangle * rectangles, int nrectangles)
{
	XFillRectangles (display, pixmap, gcp, rectangles, nrectangles);
	if (! synchro_mode)
		XFillRectangles (display, mainwindow, gc,
			rectangles, nrectangles);
}

void ScreenXWin::myXDrawSegments (XSegment * seg, int nsegments)
{
	XDrawSegments (display, pixmap, gcp, seg, nsegments);
	if (! synchro_mode)
		XDrawSegments (display, mainwindow, gc, seg, nsegments);
}


std::string check_display (const std::string & sysname)
{
	if (! sysname.empty () )
		return sysname;
	const char * strDISPLAY= getenv ("DISPLAY");
	if (strDISPLAY && strDISPLAY [0] != '\0')
		return strDISPLAY;
	else
		return std::string ();
}

GraphicsSystem * createSystemXWin (const std::string & sysname)
{
	TRF;

	return new GraphicsSystemXWin (sysname);
}



} // namespace graphics_impl

} // namespace blassic


#endif

// End of graphics_xlib.cpp
